/*!
 * @file ArpStepper.cpp
 *
 * @mainpage arpschuino lib for steppers
 *
 * @section intro_sec Introduction
 *
 * This is the documentation for arpschuino ArpStepper library. 
 *
 * @section dependencies Dependencies
 *
 * This library depends on <a href="https://github.com/NicksonYap/digitalWriteFast"> digitalWriteFast</a>
 * 
 * It is automatically installed with the arpschuino core
 *
 * @section author Author
 *
 * Written by Jacques Bouault for arpschuino.
 * 
 * @section version Version
 * 
 * version 0.1.0 adapted to ESP32
 * 
 * @section date Date
 * 
 * 2022 june 08
 *
 * @section license License
 *
 * <a href="https://en.wikipedia.org/wiki/Beerware"> Beer-ware</a> license  
 *
 */
 
#include "Arduino.h"
#include "ARPStepper.h"


/**************************************************************************/
/*!
    @brief    Constructor.Without parameter            
*/
/**************************************************************************/
ArpStepper::ArpStepper() ://Constructor

    m_inverted(false),
    m_step_per_revolution(400),//step per revolution;//nombre de pas par tour
    
    m_revolutions(10),//number of turns desired (in float to have decimal numbers)
                      //nombre de tours souhaite(en float pour avoir les nombres a virgule)
    m_impulse (20),//pulse duration in micro seconds
                    //durée de l'impulsion en micro secondes
    m_wait (100),//waiting time, in ms, in the event of a change of direction
                  //duree de l'attente,  en ms, en cas de changement de direction
    m_RPM_min(5),//minimum speed in revolutions / minute
                  //la vitesse minimale en tours/minute
    m_RPM_max(240),//la vitesse maximale en tours/minute
                    //maximum speed in revolutions / minute 
    m_dir(2),//outputs to be connected to the terminal DIR of the driver
                //les sorties a relier a la borne DIR du driver
    m_pull(3),//outputs to be connected to the terminal PULL the driver
                  //les sorties a relier a la borne PULL du driver  
  


    m_nbr_pas(0),
    m_previous_pas(0),
    m_previousMicros(0),
    m_resolution(255) ,
    m_must_refresh_pos_order(false),
    m_must_refresh_speed_order(false),
    m_val(0),
    m_val_speed(0)
{
    pinMode (m_pull, OUTPUT);
    pinMode (m_dir, OUTPUT);      
}
/**************************************************************************/
/*!
    @brief    Constructor.
              3 parameters for the hardware config.
    @param    dir
              outputs to be connected to the terminal DIR of the driver.
    @param    pull
              outputs to be connected to the terminal PULL (or STEP) of the driver.
    @param    step_per_revolution
              Steps per revolution, including the microsteps of the driver.              
*/
/**************************************************************************/
ArpStepper::ArpStepper(byte dir,byte pull, uint16_t step_per_revolution) ://Constructor

    m_inverted(false),
    m_step_per_revolution(step_per_revolution),//step per revolution;//nombre de pas par tour
    
    m_revolutions(10),//number of turns desired (in float to have decimal numbers)
                      //nombre de tours souhaite(en float pour avoir les nombres a virgule)
    m_impulse (20),//pulse duration in micro seconds
                    //durée de l'impulsion en micro secondes
    m_wait (100),//waiting time, in ms, in the event of a change of direction
                  //duree de l'attente,  en ms, en cas de changement de direction
    m_RPM_min(5),//minimum speed in revolutions / minute
                  //la vitesse minimale en tours/minute
    m_RPM_max(240),//la vitesse maximale en tours/minute
                    //maximum speed in revolutions / minute 
    m_dir(dir),//outputs to be connected to the terminal DIR of the driver
                //les sorties a relier a la borne DIR du driver
    m_pull(pull),//outputs to be connected to the terminal PULL the driver
                  //les sorties a relier a la borne PULL du driver  
  


    m_nbr_pas(0),
    m_previous_pas(0),
    m_previousMicros(0),
    m_resolution(255) ,
    m_must_refresh_pos_order(false),
    m_must_refresh_speed_order(false),
    m_val(0),
    m_val_speed(0)
{
    pinMode (m_pull, OUTPUT);
    pinMode (m_dir, OUTPUT);      
}

/**************************************************************************/
/*!
    @brief    Attach pins and define tehe number of steps per revolution.
              3 parameters for the hardware config.
    @param    dir
              outputs to be connected to the terminal DIR of the driver.
    @param    pull
              outputs to be connected to the terminal PULL (or STEP) of the driver.
    @param    step_per_revolution
              Steps per revolution, including the microsteps of the driver.              
*/
/**************************************************************************/
void ArpStepper::attach(byte dir,byte pull, uint16_t step_per_revolution) 
{
    m_step_per_revolution = step_per_revolution;//step per revolution;//nombre de pas par tour
    m_dir = dir;//outputs to be connected to the terminal DIR of the driver
                //les sorties a relier a la borne DIR du driver
    m_pull = pull;//outputs to be connected to the terminal PULL the driver
                  //les sorties a relier a la borne PULL du driver      
    pinMode (m_pull, OUTPUT);
    pinMode (m_dir, OUTPUT);      
}
/**************************************************************************/
/*!
    @brief  Initialisation.
    @param    RPM_min
              minimum speed in revolutions / minute. Default 5.
    @param    RPM_max
              maximum speed in revolutions / minute. Default 240.
    @param    revolutions
              number of turns desired (in float to have decimal numbers). Default 10.   
    @param    resolution
              The resolution of the position command. Default 255 (8bits).
              (1 to 65535) or _7bits (127), _8bits (255), _10bits (1023), _12bits (4095), _16bits (65535)                     
*/
/**************************************************************************/
void ArpStepper::init (int16_t RPM_min,int16_t RPM_max ,float revolutions,uint16_t resolution)//(RPM_min, RPM_max , revolution, resolution)//
{
    m_revolutions=revolutions;
    m_RPM_min=RPM_min;
    m_RPM_max=RPM_max;
    m_resolution=resolution;
    
    m_motor_speed_min = calcul_step_duration(m_RPM_min, m_step_per_revolution);
    Serial.println();
    Serial.print(" m_RPM_min : ");Serial.println(m_RPM_min);
    Serial.print(" m_motor_speed_min : ");Serial.println(m_motor_speed_min);
     m_motor_speed_max = calcul_step_duration(m_RPM_max, m_step_per_revolution);  
    Serial.print(" m_motor_speed_max : ");Serial.println(m_motor_speed_max); 

    m_factor=(float)(m_step_per_revolution*m_revolutions)/m_resolution;
    //m_speed_factor = (m_motor_speed_max-m_motor_speed_min); 
    m_speed_factor = (float)(m_motor_speed_max-m_motor_speed_min);
    m_speed_factor = m_speed_factor /255;
}

/**************************************************************************/
/*!
    @brief  invert the direction of rotation.
    @param    inverted
              true or false.                
*/
/**************************************************************************/
void ArpStepper::invert_rotation(bool inverted)
{
  m_inverted = inverted;
}

/**************************************************************************/
/*!
    @brief    orders the motor to reach a given position, at a given speed.
    @return 1 if a step was made forward, -1 if a step was made backward, 0 if no step was made.
*/
/**************************************************************************/
int8_t ArpStepper::perform()
{
    #ifdef __AVR__
      digitalWriteFast(m_pull,LOW);
    #else
      digitalWrite(m_pull,LOW);
    #endif    
   calcul_nbr_steps(m_val);
   set_direction();
   return do_step(m_val_speed);  
}

/**************************************************************************/
/*!
    @brief    orders the motor to turn in a direction, at a given speed.
    @param    val
              given direction.
    @param    val_speed
              speed at which to turn.
    @return 1 if a step was made, 0 if no step was made.          
*/
/**************************************************************************/
bool ArpStepper::continuous()
{
  //end_of_pulse();
  #ifdef __AVR__
    digitalWriteFast(m_pull,LOW);
  #else
    digitalWrite(m_pull,LOW);
  #endif    
  if(m_must_refresh_pos_order==true)
  {
    m_must_refresh_pos_order=false;
    bool sens;
    if(m_inverted)
    {
      if (m_val<(m_resolution/2)+1)     sens=HIGH;// + m_inverted;
      else                              sens=LOW;// - m_inverted;
    }
    else
    {
      if (m_val<(m_resolution/2)+1)     sens=LOW;// + m_inverted;
      else                              sens=HIGH;
    }
    
    if (sens != m_previous_sens)
    {
      m_change_sens = 1; 
      #ifdef __AVR__
        digitalWriteFast(m_dir,sens);
      #else
        digitalWrite(m_dir,sens);
      #endif        
    }      
    else                              
      m_change_sens = 0;
   
    m_previous_sens = sens;
  }
 
  if (m_val_speed>0)
    return step_plus(m_val_speed);    

  else 
    return 0;  
}

/**************************************************************************/
/*!
    @brief    update order datas.
    @param    val
              given position or given direction.
    @param    val_speed
              speed at which to turn.
*/
/**************************************************************************/
void ArpStepper::refresh(uint16_t val,uint16_t val_speed)
{
    m_must_refresh_pos_order=true;
    m_val=val;
    m_must_refresh_speed_order=true;
    m_val_speed=speed_curve[val_speed];
}
 

void ArpStepper::set_wait(uint16_t wait)
{
	m_wait=wait;
}
//____________________________________________________________________________________//

uint32_t ArpStepper::calcul_nbr_steps(uint32_t val)  //calcul ne nombre de pas qu'il reste a faire
{
  if(m_must_refresh_pos_order==true)
  {
    m_must_refresh_pos_order=false;
    m_position_order=val*m_factor;
    //set_direction();   
  }   
  m_nbr_pas = m_position_order - m_previous_pas;
    
  return m_nbr_pas;
}


//____________________________________________________________________________________//

bool ArpStepper::set_direction()
{
  bool sens;
  if (m_nbr_pas>0)                  sens=LOW + m_inverted;
  else                              sens=HIGH - m_inverted;
  
  if (sens != m_previous_sens)      m_change_sens = 1;
  else                              m_change_sens = 0;
  
  m_previous_sens = sens;
  #ifdef __AVR__
    digitalWriteFast(m_dir,sens);
  #else
    digitalWrite(m_dir,sens);
  #endif  
  return  m_change_sens;
}

//____________________________________________________________________________________//

bool ArpStepper::step_plus(uint16_t val_speed)
{
  if(val_speed>0)
  {
    if(m_must_refresh_speed_order)
    {
      m_must_refresh_speed_order=false;
      //m_interval = (val_speed) * (m_speed_factor) + m_motor_speed_min;//in microseconds

      m_interval =   map(val_speed,0,255,m_motor_speed_min,m_motor_speed_max);                           // (val_speed) * (m_speed_factor) + m_motor_speed_min;//in microseconds
      // m_interval = (val_speed) * (m_speed_factor) + m_motor_speed_min;//in microseconds
      // Serial.print(m_val_speed);////////////////////////////////////////debug
      // Serial.print(" * ");////////////////////////////////////////debug      
      // Serial.print(m_speed_factor);////////////////////////////////////////debug
      // Serial.print(" + ");////////////////////////////////////////debug      
      // Serial.print(m_motor_speed_min);////////////////////////////////////////debug      
      // Serial.print(" = ");////////////////////////////////////////debug
       Serial.println(m_interval);////////////////////////////////////////debug
    }  

    if(m_change_sens)//si on change de sens
    {
      m_interval += m_wait;        //une petie pause avant de repartir dans l'autre sens
      m_change_sens = 0;
    }  

    unsigned long currentMicros = micros();
    //if (currentMicros - m_previousMicros+250 >= m_interval)//250 pour compenser le temps d'execution
    if (currentMicros - m_previousMicros >= m_interval)
    {
      m_previousMicros = currentMicros;
      
      #ifdef __AVR__
        digitalWriteFast(m_pull,HIGH);
      #else
        digitalWrite(m_pull,HIGH);
      #endif      
      return 1;
    }
    else{
      return 0;
    }
  }
  else return 0;
}

//____________________________________________________________________________________//

int8_t ArpStepper::do_step(uint16_t val_speed)
{
   if(m_nbr_pas>0)
   {
        m_previous_pas += step_plus(val_speed);
        return 1;              
   }
      
   else  if(m_nbr_pas<0)
   {
        m_previous_pas -= step_plus(val_speed);
        return -1;
   }
   else
        return 0;   
}

//____________________________________________________________________________________//

uint32_t ArpStepper::calcul_step_duration(uint16_t RPM, uint16_t step_per_revolution)const
{
  float RPS = (float)RPM/60; //tours par secondes
  //Serial.println();
  //Serial.print("RPS : ");Serial.println(RPS);
  float round_duration = 1000/RPS;//la duree d'un tour, en milli-secondes
  //Serial.print("round_duration : ");Serial.println(round_duration);
  uint32_t micro_round_duration = round_duration * 1000;//la duree d'un tour, en micro-secondes
  //Serial.print("micro_round_duration : ");Serial.println(micro_round_duration);
  uint32_t step_duration = micro_round_duration/step_per_revolution;//La duree d'un pas, en micro secondes
  //Serial.print("step_duration : ");Serial.println(step_duration);

  return step_duration; 
}

 //____________________________________________________________________________________//

void ArpStepper::set_resolution(uint16_t resolution)
{
    m_resolution=resolution;
}

//____________________________________________________________________________________//
const bool ArpStepper::end_of_pulse()
{
    if (micros() - m_previousMicros >= m_impulse)
    {
      #ifdef __AVR__
        digitalWriteFast(m_pull,LOW);
       #else
        digitalWrite(m_pull,LOW);
       #endif
      return 1;      
    }
    else
      return 0;
}


////////////////////////////////////////////////////////////////////////////////////////

const uint8_t ArpStepper::speed_curve[256]{    0  ,
1 ,
3 ,
5 ,
7 ,
9 ,
11  ,
13  ,
15  ,
17  ,
19  ,
21  ,
22  ,
24  ,
26  ,
28  ,
30  ,
32  ,
34  ,
36  ,
37  ,
39  ,
41  ,
43  ,
45  ,
47  ,
48  ,
50  ,
52  ,
54  ,
55  ,
57  ,
59  ,
61  ,
62  ,
64  ,
66  ,
68  ,
69  ,
71  ,
73  ,
74  ,
76  ,
78  ,
79  ,
81  ,
83  ,
84  ,
86  ,
88  ,
89  ,
91  ,
92  ,
94  ,
96  ,
97  ,
99  ,
100 ,
102 ,
103 ,
105 ,
106 ,
108 ,
109 ,
111 ,
112 ,
114 ,
115 ,
117 ,
118 ,
120 ,
121 ,
123 ,
124 ,
126 ,
127 ,
128 ,
130 ,
131 ,
133 ,
134 ,
135 ,
137 ,
138 ,
139 ,
141 ,
142 ,
143 ,
145 ,
146 ,
147 ,
149 ,
150 ,
151 ,
152 ,
154 ,
155 ,
156 ,
157 ,
159 ,
160 ,
161 ,
162 ,
164 ,
165 ,
166 ,
167 ,
168 ,
169 ,
171 ,
172 ,
173 ,
174 ,
175 ,
176 ,
177 ,
178 ,
180 ,
181 ,
182 ,
183 ,
184 ,
185 ,
186 ,
187 ,
188 ,
189 ,
190 ,
191 ,
192 ,
193 ,
194 ,
195 ,
196 ,
197 ,
198 ,
199 ,
200 ,
201 ,
202 ,
202 ,
203 ,
204 ,
205 ,
206 ,
207 ,
208 ,
209 ,
209 ,
210 ,
211 ,
212 ,
213 ,
214 ,
214 ,
215 ,
216 ,
217 ,
217 ,
218 ,
219 ,
220 ,
220 ,
221 ,
222 ,
223 ,
223 ,
224 ,
225 ,
225 ,
226 ,
227 ,
227 ,
228 ,
229 ,
229 ,
230 ,
231 ,
231 ,
232 ,
232 ,
233 ,
234 ,
234 ,
235 ,
235 ,
236 ,
236 ,
237 ,
237 ,
238 ,
238 ,
239 ,
239 ,
240 ,
240 ,
241 ,
241 ,
242 ,
242 ,
243 ,
243 ,
244 ,
244 ,
244 ,
245 ,
245 ,
246 ,
246 ,
246 ,
247 ,
247 ,
247 ,
248 ,
248 ,
248 ,
249 ,
249 ,
249 ,
250 ,
250 ,
250 ,
250 ,
251 ,
251 ,
251 ,
251 ,
252 ,
252 ,
252 ,
252 ,
253 ,
253 ,
253 ,
253 ,
253 ,
253 ,
254 ,
254 ,
254 ,
254 ,
254 ,
254 ,
254 ,
254 ,
255 ,
255 ,
255 ,
255 ,
255 ,
255 ,
255 ,
255 ,
255 ,
255 ,
255}; 
